/* -------------------------------------------------------------------------- **
** PURPOSE:  Solve a set of first order ordinary differential equations of    **
**           the form dy(i)/dt = F(t,y(1), ..., y(numY) (i = 1, ..., numY).   **
**                                                                            **
** INPUT:                                                                     **
**     eqns: Function that evaluates dy(i)/dt (i = 1, ..., numY), the         **
**           first derivatives of y(1), ..., y(numY) with respect to t.       **
**                                                                            **
**     numY: The number of differential equations to be solved.               **
**        y: One-dimensional array whose elements are y(1), ..., y(numY).     **
**                                                                            **
**   tStart: Value of independent variable t on entry to this function.       **
**                                                                            **
**   hEntry: Integration stepsize upon entry to this function.                **
**           hEntry may be NULL if calcDeriv0OrIntegrate = 0.                 **
**           On exit, hEntry is set to actual stepsize used by integrator.    **
**                                                                            **
** absError: Allowable absolute error in y(i)  (i=1, ..., numY).              **
** relError: Allowable relative error in y(i)  (i=1, ..., numY).              **
**                                                                            **
** calcDeriv0OrIntegrate:                                                     **
**           When calcDeriv0OrIntegrate = 0, the function eqns is             **
**           called and dy(i)/dt (i = 1, ..., numY) are evaluated,            **
**           but no integration is performed.                                 **
**                                                                            **
**           When calcDeriv0OrIntegrate = 1, the Kutta-Merson algorithm       **
**           (L. Fox, Numerical Solutions of Ordinary and Partial Differ-     **
**           ential Equations, Palo Alto: Addison-Wesley, 1962, pp. 24-25)    **
**           is employed to perform the numerical solution of the differ-     **
**           ential equations - e.g., from tStart to tStart+deltaT.           **
**                                                                            **
** OUTPUT:   The actual stepsize used by integrator is returned in hEntry.    **
**           The values of y(i) (i=1 ... numY) at t+hEntry are returned in y. **
**           The return value is the integrator's estimate of the next step-  **
**           size or 0 if an error was encountered and integration stopped.   **
**                                                                            **
** SOURCE:   Copyright (c) 1988-2015 by Paul Mitiguy.  All rights reserved.   **
**           Only Motion Genesis currently-paid-up professional and research  **
**           licensees may use this code.  No other use (or distribution)     **
**           permitted. Currently paid-up professional and research licensees **
**           are granted the right to copy, modify, and distribute this code  **
**           provided this copyright notice appears in all copies and distri- **
**           butions AND licensee continuously pays its license/maintenance   **
**           fees for the duration of use by licensee or its distributees.    **
**           This copyright also applies to the functions:                    **
**           KMIntegrator, IntegratorStep, IntegrateForwardOrBackward         **
** -------------------------------------------------------------------------- */
static double  KMIntegrator( char*(*eqns)(double, double*, double*, char),
                             int numY, double y[], double tStart, double *hEntry, double smallestAllowableStepsize,
                             double absError, double relError, char calcDeriv0OrIntegrate )
{
   double f0[_NumberOfODES_], f1[_NumberOfODES_], f2[_NumberOfODES_], y1[_NumberOfODES_], y2[_NumberOfODES_];
   double h = hEntry ? *hEntry : 0;              /* Dereferenced local copy */
   int   errorInVarIndex = 0;                    /* Variable that failed    */
   char *errorMessageEqns = NULL;
   if( numY > _NumberOfODES_ )  { printf("\nError: Increase _NumberOfODES_ in KMIntegrator.\n" );  return 0; }

   /* Always need to calculate derivatives at tStart (always boundary here) */
   /* If calcDeriv0OrIntegrate == 0,  just call eqns at tStart and return.  */
   errorMessageEqns = (*eqns)( tStart, y, f0, 1 );  /* Integration boundary */
   if( !errorMessageEqns && calcDeriv0OrIntegrate == 0 )  return 1;  

   while( tStart+h != tStart && !errorMessageEqns )      /* Avoid round-off */
   {
      double errorRatioMax = 0;                  /* Testing error criterion */
      double h2 = h * 0.5;                                 /* Half    of h  */
      double h3 = h / 3.0;                                 /* Third   of h  */
      double h6 = h / 6.0;                                 /* Sixth   of h  */
      double h8 = h * 0.125;                               /* Eighth  of h  */
      int i;
      for( i=0;  i<numY;  i++ )  y1[i] = y[i] + h3*f0[i];
      if( (errorMessageEqns = (*eqns)( tStart+h3, y1, f1, 0 )) != NULL ) break;
      for( i=0;  i<numY;  i++ )  y1[i] = y[i] + h6*(f0[i] + f1[i]);
      if( (errorMessageEqns = (*eqns)( tStart+h3, y1, f1, 0 )) != NULL ) break;
      for( i=0;  i<numY;  i++ )  y1[i] = y[i] + h8*(f0[i] + 3*f1[i]);
      if( (errorMessageEqns = (*eqns)( tStart+h2, y1, f2, 0 )) != NULL ) break;
      for( i=0;  i<numY;  i++ )  y1[i] = y[i] + h2*(f0[i] - 3*f1[i] + 4*f2[i]);
      if( (errorMessageEqns = (*eqns)( tStart+h,  y1, f1, 0 )) != NULL)  break;
      for( i=0;  i<numY;  i++ )  y2[i] = y[i] + h6*(f0[i] + 4*f2[i] + f1[i]);

      /* Both y1[i] and y2[i] provide estimates to the new value of y.      */
      /* Decide if the relative and absolute error tolerances are met.      */ 
      /* If they are not, reduce the stepsize and restart the integration.  */
      for( i=0;  i<numY;  i++ )                  /* Check all variables     */
      {
         double errorInEstimate = 0.2 * fabs( y2[i] - y1[i] );   
         double relTest  = fabs( y2[i] ) * relError; 
         double largerOfAbsErrOrRelTest = absError > relTest ? absError : relTest;
         double errorRatio = errorInEstimate / largerOfAbsErrOrRelTest;
         if( errorRatio > errorRatioMax ) { errorRatioMax = errorRatio;  errorInVarIndex = i; }
      }

      /* If the errorRatioMax >= 1, either the absError or relTest failed.  */
      /* In other words, errorInEstimate >= largerOfAbsErrOrRelTest         */ 
      if( errorRatioMax >= 1 )  
      {                                          /* Halve stepsize & restart*/
         if( fabs(h=h2) > fabs(smallestAllowableStepsize) )  continue;
         printf( "\n Error: Stepsize has been cut too many times.\n" );  break;
      }

      /* Integration succeeded.  Update the values of y and t before return */
      for( i = 0;  i < numY;  i++ )  y[i] = y2[i];   

      /* Return actual stepsize and estimate for next integration stepsize. */
      *hEntry = h;   return errorRatioMax < 1.0/64.0 ? h+h : h;
   }

   /* Print error messages since numerical integrator failed. */
   if( errorMessageEqns ) printf("\n %s\n Error: Numerical integration failed at t = %15.8E.\n" , errorMessageEqns, tStart );
   else 
   {
      printf("\n Error: Numerical integration of variable %d failed at t = %15.8E." , errorInVarIndex, tStart );
      printf("\n Error: Numerical round-off makes stepsize of %15.7E too small relative to %15.7E."
             "\n        Numerical integration stepsize may have been cut too many times.\n\n", h, tStart );
   }
   (*eqns)( tStart, y, f0, 1 );                  /* Fill for error display. */
   return hEntry ? (*hEntry = 0) : 0;
}


/* -------------------------------------------------------------------------- **
** PURPOSE:  Controls the independent variable, e.g., t while solving a set   **
**           of first order ordinary differential equations of the form       **
**           dy(i)/dt = F(t,y(1), ..., y(numY) (i = 1, ..., numY).            ** 
**           Discontinuities in functions and t should be handled here.       **
**                                                                            **
** INPUT:    eqns, y, t, numY, absError, relError, calcDeriv0OrIntegrate are  **
**           described above.  If numY = 0, the static (stored) value of the  **
**           integration stepsize is reset to tStep.  This is useful when     **
**           integrating several sets of different differential equations.    **
**                                                                            **
**      t:   Independent variable.                                            **
**  tStep:   Maximum integration stepsize.                                    **
**                                                                            **
** OUTPUT:   The value of  t + tStep  is returned in t.                       **
**           The values of y(i) at  t + tStep  are returned in y.             **
**           The return value is 0 if the algorithm fails.                    **
** -------------------------------------------------------------------------- */
static int  IntegratorStep( char*(*eqns)(double, double*, double*, char),
                            int numY, double y[], double *t, double tStep, 
                            double absError, double relError, char calcDeriv0OrIntegrate )
{
   static double hc = 0.0;                       /* Last value of stepsize  */
   double h;                                     /* Current stepsize        */
   double hAccumulated = 0;                      /* How far to tStep        */
   double smallestAllowableStepsize = 1.0E-7 * tStep;

   /* Decide if the static (stored) value of hc needs to be initialized.    */
   if( numY == 0 ) { hc = tStep;  return 1; }    /* (Re)-initialize hc      */

   /* If calcDeriv0OrIntegrate = 0, just call eqns at exactly *t            */
   if( calcDeriv0OrIntegrate == 0 )  return (int)KMIntegrator( eqns, numY, y, *t, NULL, 0, 0, 0, 0 );

   /* Upon entry, initialize the current stepsize to the stored value of hc */
   if( tStep == 0 || absError==0 ) return 0;     /* Cannot integrate        */
   if( hc * tStep < 0 )  hc = -hc;               /* Switch direction        */
   else if( hc == 0 )  hc = tStep;               /* Maybe initial entry     */
   if( fabs(hc) > fabs(tStep) ) hc = tStep;      /* Ensure hc not too big   */
   h  = hc;                                      /* Current stepsize        */

   /* Make as many little integration steps as necessary to get to end      */
   /* Each integration step starts at *t and ends at *t + h.                */
   while( *t+h != *t )                           /* Check round-off problems*/
   {
      /* Numerically integrate y[i] and maybe get shorter value of h.       */
      /* Set hNext to integrator's estimate of the next integration step-   */
      /* size or 0 if an error was encountered and integration stopped.     */
      double hToIntegrator = h;                  /* Suggested stepsize.     */
      double hNext = KMIntegrator( eqns, numY, y, *t, &h, smallestAllowableStepsize, absError, relError, calcDeriv0OrIntegrate );
      char integrationFinished = fabs( (hAccumulated+=h) + smallestAllowableStepsize) > fabs(tStep);
      if( h==0 || hNext==0 ) return 0;           /* Integration failed      */
      
      /* Any function or time discontinuities should be handled here.       */
      /* Increment value of t using stepsize returned by the integrator.    */
      *t += h;                                   /* Prepare for next step.  */

      /* Update the static (stored) value of hc for next call to integrator.*/
      /* Always change the stepsize if not taking the last step.            */
      /* Always reduce the stepsize if the integrator reduced the stepsize. */
      /* Could always increase stepsize if integrator suggests larger one,  */
      /* e.g., by adding   || fabs(hNext) > fabs(hc)   to logic below.      */
      if( !integrationFinished || fabs(hNext) < fabs(hToIntegrator) || fabs(h) < fabs(hToIntegrator) )  h = hc = hNext; 

      /* If finished, calculate derivatives of y at the final value of *t.  */
      if( integrationFinished ) return (int)KMIntegrator( eqns, numY, y, *t, NULL, 0, 0, 0, 0 );
      
      /* Integrator may suggest a stepsize that is bigger than allowed.     */
      /* If next jump is past or very close to end of tStep, adjust h.      */
      if( fabs(hAccumulated + 1.1*h) > fabs(tStep) )  h = tStep - hAccumulated;
   }
   printf("\n Error: Stepsize of %15.7E is too small relative to %15.7E."
          "\n Numerical integration halted at t = %15.8E due to numerical round-off."
          "\n The stepsize may have been cut too many times.\n\n", h, *t, *t );
   return 0;
}


/* ------------------------------------------------------------------------ */
static int  IntegrateForwardOrBackward( int numVariables, double varArrayToIntegrate[], double outputToFill[], double tInitial, double tFinal, double tStep, double absError, double relError, unsigned int printIntScreen, unsigned int printIntFile )
{
   double t = tInitial,  tFinalMinusEpsilon = tFinal - 0.0001*tStep;
   char   integrateForward = (tFinal >= tInitial),  calcDeriv0OrIntegrate = 0;
   int    printCounterScreen = 0,  printCounterFile = 0;

   /* Initialize numerical integrator with call at t = tInitial, thereafter integrate. */
   while( 1 )
   {
      int integrationOK = IntegratorStep( eqns1, numVariables, varArrayToIntegrate, &t, tStep, absError, relError, calcDeriv0OrIntegrate );  
      int isIntegrationFinished = !integrationOK || (integrateForward && t > tFinalMinusEpsilon) || (!integrateForward && t < tFinalMinusEpsilon);
      int shouldPrintToScreen = printIntScreen && ( isIntegrationFinished || --printCounterScreen <= 0 );
      int shouldPrintToFile   = printIntFile   && ( isIntegrationFinished || --printCounterFile   <= 0 );
      if( shouldPrintToScreen || shouldPrintToFile || isIntegrationFinished ) CalculateOutput( t, outputToFill );
      if( shouldPrintToScreen || shouldPrintToFile )
      {
         OutputToScreenOrFile( outputToFill, t, integrationOK, shouldPrintToScreen, shouldPrintToFile );
         if( shouldPrintToScreen ) printCounterScreen = (int)printIntScreen; 
         if( shouldPrintToFile )   printCounterFile   = (int)printIntFile;  
      }
      if( isIntegrationFinished ) return integrationOK;
      calcDeriv0OrIntegrate = 1;
   }
   return 1;
}

