/* -------------------------------------------------------------------------- **
** PURPOSE:  The matrix equation A*X = B is solved for X, where A is an       **
**           n by n matrix, and X and B are n by 1 matrices.                  **
**                                                                            **
** INPUT:    numberOfEqns: The integer n (number of equations).               **
**           A: An array of n pointers to rows of doubles.                    **
**              In effect, A is an n by n array of double precision numbers.  **
**           B: An n by 1 array of double precision numbers.                  **
**              The elements of B are unchanged on return.                    **
**                                                                            **
** OUTPUT:   Returns through X an n by 1 array of double precision numbers.   **
**           Returns 0 if no solution could be found.                         **
**                                                                            **
** SOURCE:   Copyright (c) 1988-2015 by Paul Mitiguy.  All rights reserved.   **
**           Only Motion Genesis currently-paid-up professional and research  **
**           licensees may use this code.  No other use (or distribution)     **
**           permitted. Currently paid-up professional and research licensees **
**           are granted the right to copy, modify, and distribute this code  **
**           provided this copyright notice appears in all copies and distri- **
**           butions AND licensee continuously pays its license/maintenance   **
**           fees for the duration of use by licensee or its distributees.    **
** -------------------------------------------------------------------------- */
static int  SolveLinearEquation( int numberOfEqns, double *A[], double B[], double X[] )
{
   static double rowScaleFactor[ _NumberOfCoupledLinearEqns_ ]; 
   double        pivotReciprocal;
   int           i, j, k;
                 
   /* Ensure there is sufficient space for row-scaling factors */
   if( numberOfEqns > _NumberOfCoupledLinearEqns_ ) { printf("\n Error: Increase _NumberOfCoupledLinearEqns_ in SolveLinearEquation.\n" );  return 0; }

   /* Begin decomposition */
   for( i=0;  i<numberOfEqns;  i++ )   
   {
      double *Ai = A[i];                         /* Short-cut to this row   */

      /* Find the element in each row with the maximum absolute value */
      double rowMax = 0.0;                       /* Max absolute column     */
      for( j=0;  j<numberOfEqns;  j++ )          /* Check for zero row      */
        if( rowMax < fabs(Ai[j]) )  rowMax = fabs(Ai[j]);

      /* Issue error if row of zeros are found */
      if( rowMax == 0.0 )                     
        { printf("\n Error singular matrix: All elements in row %d of Coefficient matrix are zeros.\n", i+1);  return 0; }

      /* Keep track of row scaling factor */
      rowScaleFactor[i] = 1.0 / rowMax;                      

      /* Keep the B matrix unchanged by copying its elements into x */
      X[i] = B[i];                          
   }

   /* Until 1 column remains */
   for( k=0;  k<numberOfEqns-1;  k++ )                
   {
      double largestPivot = 0.0;                 /* Largest relative pivot  */
      int    swapi = k;
      for( i=k;  i<numberOfEqns;  i++ )          /* Check remaining rows    */
      {                                        
         double relSizeThisColumn = fabs(A[i][k]) * rowScaleFactor[i];
         if( relSizeThisColumn > largestPivot ) { swapi = i;  largestPivot = relSizeThisColumn; }  
      }                        

      /* Issue warning if zero pivot is encountered */
      if( largestPivot == 0.0 )  { printf("\n Error singular matrix: Zero pivot in column %d during LU-decomposition of COEF matrix.\n", k+1 );  return 0; }

      /* Maybe switch rows of A and x by changing row pointers and x values */
      if( swapi != k )                         
      {                                        
         double *swapa = A[k],  swapx = X[k],  swapScale = rowScaleFactor[k];
         A[k] = A[swapi];  A[swapi] = swapa;     /* Change row pointers of A*/
         X[k] = X[swapi];  X[swapi] = swapx;     /* Change row values of X  */
         rowScaleFactor[k] = rowScaleFactor[swapi];  rowScaleFactor[swapi] = swapScale;
      }

      /* Calculate value of pivot and change lower rows */
      pivotReciprocal = 1.0 / A[k][k];           
      for( i=k+1;  i<numberOfEqns;  i++ )                   
      {
         double *Ak = A[k],  *Ai = A[i];         /* Shortcuts               */
         double ratio = Ai[k] * pivotReciprocal; /* Multiplicative factor   */
         X[i] -= ratio * X[k];                   /* Modify elements of X    */
         for( j=k+1;  j<numberOfEqns;  j++ )  Ai[j] -= ratio * Ak[j];  
      }
   }

   /* Check last pivot */
   if( A[numberOfEqns-1][numberOfEqns-1] == 0.0 )  { printf("\n Error singular matrix: Zero pivot in column %d during LU-decomposition of COEF matrix.\n", numberOfEqns );  return 0; }

   /* Solve upper * X = Z  */
   for( i=numberOfEqns-1;  i>=0;  i-- )  
   {
      double sum = 0.0, *Ai = A[i];
      for( j=i+1;  j<numberOfEqns;  j++ )  sum += Ai[j] * X[j];
      X[i] = (X[i] - sum) / Ai[i];
   }
   return 1;
}

